# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-10-05 12:45
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('modoboa_contacts', '0004_auto_20181005_1415'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='contact',
            name='user',
        ),
        migrations.AlterField(
            model_name='contact',
            name='addressbook',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='modoboa_contacts.AddressBook'),
        ),
    ]



# vim: set ts=4 sw=4 expandtab sts=4:
# Copyright (c) 2011-2014 Christian Geier & contributors
# Copyright (c) 2018 Antoine Nguyen <tonio@ngyn.org>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
contains the class PyCardDAv and some associated functions and definitions
"""

from collections import namedtuple
import logging

from six.moves import urllib

from caldav import elements
import lxml.etree as ET
import requests


class UploadFailed(Exception):
    """uploading the card failed"""
    pass


class NoWriteSupport(Exception):
    """write support has not been enabled"""
    pass


nsmap = {
    "D": "DAV:",
    "CARD": "urn:ietf:params:xml:ns:carddav",
}


def ns(prefix, tag=None):
    name = "{%s}" % nsmap[prefix]
    if tag is not None:
        name = "%s%s" % (name, tag)
    return name


class Mkcol(elements.base.BaseElement):
    tag = ns("CARD", "mkcol")


class AddressBook(elements.base.BaseElement):
    tag = ns("CARD", "addressbook")


class AddressBookQuery(elements.base.BaseElement):
    tag = ns("CARD", "addressbook-query")


class AddressData(elements.base.BaseElement):
    tag = ns("CARD", "address-data")


class Etag(elements.base.BaseElement):
    tag = ns("D", "getetag")


class SyncToken(elements.base.BaseElement):
    tag = ns("D", "sync-token")


class SyncLevel(elements.base.BaseElement):
    tag = ns("D", "sync-level")


class SyncCollectionQuery(elements.base.BaseElement):
    tag = ns("D", "sync-collection")


class PyCardDAV(object):
    """class for interacting with a CardDAV server

    Since PyCardDAV relies heavily on Requests [1] its SSL verification is also
    shared by PyCardDAV [2]. For now, only the *verify* keyword is exposed
    through PyCardDAV.

    [1] http://docs.python-requests.org/
    [2] http://docs.python-requests.org/en/latest/user/advanced/

    raises:
        requests.exceptions.SSLError
        requests.exceptions.ConnectionError
        more requests.exceptions depending on the actual error
        Exception (shame on me)

    """

    def __init__(self, resource, debug='', user='', passwd='',
                 verify=True, write_support=False, auth='basic'):
        # shutup urllib3
        urllog = logging.getLogger('requests.packages.urllib3.connectionpool')
        urllog.setLevel(logging.CRITICAL)
        urllog = logging.getLogger('urllib3.connectionpool')
        urllog.setLevel(logging.CRITICAL)

        # activate pyopenssl if available
        try:
            import urllib3.contrib.pyopenssl
        except ImportError:
            pass
        else:
            urllib3.contrib.pyopenssl.inject_into_urllib3()

        split_url = urllib.parse.urlparse(resource)
        url_tuple = namedtuple('url', 'resource base path')
        self.url = url_tuple(resource,
                             split_url.scheme + '://' + split_url.netloc,
                             split_url.path)
        self.debug = debug
        self.session = requests.session()
        self.write_support = write_support
        self._settings = {'verify': verify}
        if auth == 'basic':
            self._settings['auth'] = (user, passwd,)
        if auth == 'digest':
            from requests.auth import HTTPDigestAuth
            self._settings['auth'] = HTTPDigestAuth(user, passwd)
        self._default_headers = {"User-Agent": "pyCardDAV"}

        headers = self.headers
        headers['Depth'] = '1'
        response = self.session.request('OPTIONS',
                                        self.url.resource,
                                        headers=headers,
                                        **self._settings)
        response.raise_for_status()  # raises error on not 2XX HTTP status code
        if 'addressbook' not in response.headers.get('DAV', ''):
            raise Exception("URL is not a CardDAV resource")

    @property
    def verify(self):
        """gets verify from settings dict"""
        return self._settings['verify']

    @verify.setter
    def verify(self, verify):
        """set verify"""
        self._settings['verify'] = verify

    @property
    def headers(self):
        """returns the headers"""
        return dict(self._default_headers)

    def _check_write_support(self):
        """checks if user really wants his data destroyed"""
        if not self.write_support:
            raise NoWriteSupport

    def create_abook(self):
        """Create a new address book."""
        self._check_write_support()
        data = (
            Mkcol() + [
                elements.dav.Prop() + [
                    elements.dav.ResourceType() +
                    elements.dav.Collection() +
                    AddressBook()
                ]
            ]
        )
        body = ET.tostring(data.xmlelement(), encoding="utf-8",
                           xml_declaration=True)
        response = self.session.request(
            'MKCOL', self.url.resource, data=body, headers=self.headers,
            **self._settings)
        response.raise_for_status()

    def get_abook(self):
        """does the propfind and processes what it returns

        :rtype: list of hrefs to vcards
        """
        xml = self._get_xml_props()
        abook = self._process_xml_props(xml)
        return abook

    def get_vcard(self, href):
        """
        pulls vcard from server

        :returns: vcard
        :rtype: string
        """
        response = self.session.get(self.url.base + href,
                                    headers=self.headers,
                                    **self._settings)
        response.raise_for_status()
        return response.content

    def get_sync_token(self):
        """Retrieve the current sync token."""
        headers = self.headers
        headers['Depth'] = '0'
        data = elements.dav.Propfind() + [elements.dav.Prop() + SyncToken()]
        body = ET.tostring(data.xmlelement(), encoding="utf-8",
                           xml_declaration=True)
        response = self.session.request('PROPFIND',
                                        self.url.resource,
                                        data=body,
                                        headers=headers,
                                        **self._settings)
        response.raise_for_status()
        abook = self._process_xml_props(response.content)
        return abook["sync-token"]

    def sync_vcards(self, sync_token=None):
        """Synchronize vcards using webdav sync."""
        data = (
            SyncCollectionQuery() +
            [elements.dav.Prop() + Etag()] +
            SyncToken(value=sync_token) +
            SyncLevel(value="1")
        )
        body = ET.tostring(data.xmlelement(), encoding="utf-8",
                           xml_declaration=True)
        response = self.session.request(
            "REPORT", self.url.resource, data=body, headers=self.headers,
            **self._settings)
        response.raise_for_status()
        return self.__process_report_result(response.content)

    def update_vcard(self, card, uid, etag):
        """
        pushes changed vcard to the server
        card: vcard as unicode string
        etag: str or None, if this is set to a string, card is only updated if
              remote etag matches. If etag = None the update is forced anyway
        """
        # TODO what happens if etag does not match?
        self._check_write_support()
        remotepath = str(self.url.resource + "/" + uid)
        headers = self.headers
        headers['content-type'] = 'text/vcard'
        if etag is not None:
            headers['If-Match'] = etag
        response = self.session.put(remotepath, data=card, headers=headers,
                                    **self._settings)
        if not response.ok:
            response.raise_for_status()
        # Now retrieve the new etag
        data = AddressBookQuery() + [elements.dav.Prop() + Etag()]
        body = ET.tostring(data.xmlelement(), encoding="utf-8",
                           xml_declaration=True)
        response = self.session.request(
            'REPORT', remotepath, data=body, headers=self.headers,
            **self._settings)
        response.raise_for_status()
        return self.__process_report_result(response.content)

    def delete_vcard(self, uid, etag):
        """deletes vcard from server

        deletes the resource at href if etag matches,
        if etag=None delete anyway
        :param uid: UID of card to be deleted
        :type uid: str()
        :param etag: etag of that card, if None card is always deleted
        :type href: str()
        :returns: nothing
        """
        # TODO: what happens if etag does not match, url does not exist etc ?
        self._check_write_support()
        remotepath = str(self.url.resource + "/" + uid)
        headers = self.headers
        headers['content-type'] = 'text/vcard'
        if etag is not None:
            headers['If-Match'] = etag
        response = self.session.delete(remotepath,
                                       headers=headers,
                                       **self._settings)
        response.raise_for_status()

    def upload_new_card(self, uid, card):
        """
        upload new card to the server

        :param card: vcard to be uploaded
        :type card: unicode
        :rtype: tuple of string (path of the vcard on the server) and etag of
                new card (string or None)
        """
        self._check_write_support()
        card = card.encode('utf-8')
        remotepath = str(self.url.resource + "/" + uid)
        headers = self.headers
        headers['content-type'] = 'text/vcard'  # TODO perhaps this should
        # be set to the value this carddav server uses itself
        headers['If-None-Match'] = '*'
        response = requests.put(remotepath, data=card, headers=headers,
                                **self._settings)
        if response.ok:
            parsed_url = urllib.parse.urlparse(remotepath)
            etag = response.headers.get('etag', '')
            return (parsed_url.path, etag)
        response.raise_for_status()

    def __process_report_result(self, xml):
        """Parse REPORT query result and return a dict."""
        result = {"cards": []}
        ns = "{DAV:}"
        element = ET.XML(xml)
        for item in element.iterchildren():
            if item.tag != ns + "response":
                result["token"] = item.text
                continue
            card = {}
            for d in item.iterdescendants():
                if d.tag == ns + "href":
                    card["href"] = d.text
                elif d.tag == ns + "status":
                    card["status"] = d.text
                elif d.tag == ns + "getetag":
                    card["etag"] = d.text
            result["cards"].append(card)
        return result

    def _get_xml_props(self):
        """PROPFIND method

        gets the xml file with all vcard hrefs

        :rtype: str() (an xml file)
        """
        headers = self.headers
        headers['Depth'] = '0'
        response = self.session.request('PROPFIND',
                                        self.url.resource,
                                        headers=headers,
                                        **self._settings)
        response.raise_for_status()
        return response.content

    @classmethod
    def _process_xml_props(cls, xml):
        """processes the xml from PROPFIND, listing all vcard hrefs

        :param xml: the xml file
        :type xml: str()
        :rtype: dict() key: href, value: etag
        """
        namespace = "{DAV:}"

        element = ET.XML(xml)
        abook = dict()
        for response in element.iterchildren():
            if response.tag != namespace + "response":
                continue
            href = ""
            etag = ""
            insert = False
            for refprop in response.iterchildren():
                if (refprop.tag == namespace + "href"):
                    href = refprop.text
                for prop in refprop.iterchildren():
                    for props in prop.iterchildren():
                        # different servers give different getcontenttypes:
                        # e.g.:
                        #  "text/vcard"
                        #  "text/x-vcard"
                        #  "text/x-vcard; charset=utf-8"
                        #  "text/directory;profile=vCard"
                        #  "text/directory"
                        #  "text/vcard; charset=utf-8"  CalendarServer
                        if (props.tag == namespace + "getcontenttype" and
                            props.text.split(';')[0].strip() in
                                ['text/vcard', 'text/x-vcard']):
                            insert = True
                        if (props.tag == namespace + "resourcetype" and
                            namespace + "collection" in
                                [c.tag for c in props.iterchildren()]):
                            insert = False
                            break
                        if props.tag == namespace + "getetag":
                            etag = props.text
                            break
                        if props.tag == namespace + "sync-token":
                            abook["sync-token"] = props.text
                    if insert:
                        abook[href] = etag
        return abook



"""Contacts viewsets."""

import django_filters.rest_framework
from rest_framework import decorators, filters, response, viewsets
from rest_framework.permissions import IsAuthenticated

from . import models
from . import serializers
from . import tasks


class AddressBookViewSet(viewsets.GenericViewSet):
    """Address book viewset."""

    permission_classes = (IsAuthenticated, )

    @decorators.list_route(methods=["get"])
    def default(self, request, *args, **kwargs):
        """Return default user address book."""
        abook = request.user.addressbook_set.first()
        if not abook:
            return response.Response(status_code=404)
        serializer = serializers.AddressBookSerializer(abook)
        return response.Response(serializer.data)

    @decorators.list_route(methods=["get"])
    def sync_to_cdav(self, request, *args, **kwargs):
        """Synchronize address book with CardDAV collection."""
        abook = request.user.addressbook_set.first()
        if request.user.parameters.get_value("enable_carddav_sync"):
            tasks.push_addressbook_to_carddav(request, abook)
        return response.Response({})

    @decorators.list_route(methods=["get"])
    def sync_from_cdav(self, request, *args, **kwargs):
        """Synchronize from CardDAV address book."""
        abook = request.user.addressbook_set.first()
        if not abook.last_sync:
            return response.Response()
        if request.user.parameters.get_value("enable_carddav_sync"):
            tasks.sync_addressbook_from_cdav(request, abook)
        return response.Response({})


class CategoryViewSet(viewsets.ModelViewSet):
    """Category ViewSet."""

    permission_classes = [IsAuthenticated]
    serializer_class = serializers.CategorySerializer

    def get_queryset(self):
        """Filter based on current user."""
        qset = models.Category.objects.filter(user=self.request.user)
        return qset.select_related("user")


class ContactFilter(django_filters.rest_framework.FilterSet):
    """Filter for Contact."""

    category = django_filters.CharFilter(name="categories__name")

    class Meta:
        model = models.Contact
        fields = ["categories"]


class ContactViewSet(viewsets.ModelViewSet):
    """Contact ViewSet."""

    filter_backends = [
        filters.SearchFilter,
        django_filters.rest_framework.DjangoFilterBackend]
    filter_class = ContactFilter
    permission_classes = [IsAuthenticated]
    search_fields = ("^first_name", "^last_name", "^emails__address")
    serializer_class = serializers.ContactSerializer

    def get_queryset(self):
        """Filter based on current user."""
        qset = models.Contact.objects.filter(
            addressbook__user=self.request.user)
        return qset.prefetch_related(
            "categories", "emails", "phone_numbers")

    def perform_destroy(self, instance):
        """Also remove cdav contact."""
        if self.request.user.parameters.get_value("enable_carddav_sync"):
            tasks.delete_contact_cdav(self.request, instance)
        instance.delete()


class EmailAddressViewSet(viewsets.ReadOnlyModelViewSet):
    """EmailAddress viewset."""

    filter_backends = [filters.SearchFilter]
    permission_classes = [IsAuthenticated]
    search_fields = (
        "^address", "^contact__display_name",
        "^contact__first_name", "^contact__last_name",
    )
    serializer_class = serializers.EmailAddressWithNameSerializer

    def get_queryset(self):
        """Filter based on current user."""
        return models.EmailAddress.objects.filter(
            contact__addressbook__user=self.request.user)



"""Contacts factories."""

import factory

from modoboa.admin import factories as admin_factories

from . import models


class AddressBookFactory(factory.django.DjangoModelFactory):
    """Address book factory."""

    class Meta:
        model = models.AddressBook

    user = factory.SubFactory(admin_factories.UserFactory)
    name = "Contacts"
    _path = "contacts"


class CategoryFactory(factory.django.DjangoModelFactory):
    """Category factory."""

    class Meta:
        model = models.Category


class EmailAddressFactory(factory.django.DjangoModelFactory):
    """Email address factory."""

    class Meta:
        model = models.EmailAddress

    type = "home"


class PhoneNumberFactory(factory.django.DjangoModelFactory):
    """Phone number factory."""

    class Meta:
        model = models.PhoneNumber

    type = "home"


class ContactFactory(factory.django.DjangoModelFactory):
    """Contact factory."""

    class Meta:
        model = models.Contact

    first_name = "Homer"
    last_name = "Simpson"
    display_name = factory.LazyAttribute(
        lambda c: "{}{}".format(c.first_name, c.last_name))

    @factory.post_generation
    def categories(self, create, extracted, **dummy_kwargs):
        """Add categories to contact."""
        if not create or not extracted:
            return
        for item in extracted:
            self.categories.add(item)

    @factory.post_generation
    def emails(self, create, extracted, **dummy_kwargs):
        """Add emails to contact."""
        if not create or not extracted:
            return
        for item in extracted:
            EmailAddressFactory(contact=self, address=item)

    @factory.post_generation
    def phone_numbers(self, create, extracted, **dummy_kwargs):
        """Add phone numbers to contact."""
        if not create or not extracted:
            return
        for item in extracted:
            PhoneNumberFactory(contact=self, number=item)



from __future__ import unicode_literals

from django.apps import AppConfig


class ModoboaContactsConfig(AppConfig):
    name = "modoboa_contacts"

    def ready(self):
        from . import handlers



# coding: utf-8
"""Contacts backend tests."""

import httmock

from django import forms
from django.urls import reverse
from django.utils import timezone

from modoboa.admin import factories as admin_factories
from modoboa.core import models as core_models
from modoboa.lib.tests import ModoAPITestCase, ModoTestCase
from modoboa.parameters import forms as param_forms
from modoboa.parameters import tools as param_tools

from . import factories
from . import mocks
from . import models


class RadicaleParametersForm(param_forms.AdminParametersForm):
    """Since contacts plugin depends on radicale, we need this."""

    app = "modoboa_radicale"

    server_location = forms.URLField(
        label="Server URL",
        help_text=(
            "The URL of your Radicale server. "
            "It will be used to construct calendar URLs."
        ),
        widget=forms.TextInput(attrs={"class": "form-control"})
    )


param_tools.registry.add("global", RadicaleParametersForm, "Radicale")


class TestDataMixin(object):
    """Create some data."""

    @classmethod
    def setUpTestData(cls):
        """Create test data."""
        super(TestDataMixin, cls).setUpTestData()
        admin_factories.populate_database()
        cls.user = core_models.User.objects.get(username="user@test.com")
        cls.addressbook = cls.user.addressbook_set.first()
        cls.category = factories.CategoryFactory(user=cls.user, name="Family")
        cls.contact = factories.ContactFactory(
            addressbook=cls.addressbook, emails=["homer@simpson.com"],
            phone_numbers=["01234567889"],
        )
        factories.ContactFactory(
            addressbook=cls.addressbook,
            first_name="Marge", emails=["marge@simpson.com"],
            categories=[cls.category]
        )
        factories.ContactFactory(
            addressbook=cls.addressbook,
            first_name="Bart", emails=["bart@simpson.com"])

    def setUp(self):
        """Initiate test context."""
        self.client.force_login(self.user)
        self.set_global_parameter(
            "server_location", "http://example.test/radicale/",
            app="modoboa_radicale")

    def enable_cdav_sync(self):
        """Enable sync. for user."""
        url = reverse("core:user_preferences")
        with httmock.HTTMock(mocks.options_mock, mocks.mkcol_mock):
            response = self.client.post(
                url, {"modoboa_contacts-enable_carddav_sync": True,
                      "modoboa_contacts-sync_frequency": 300},
            )
        self.assertEqual(response.status_code, 200)


class ViewsTestCase(TestDataMixin, ModoTestCase):
    """Check views."""

    def test_user_settings(self):
        """Check that remote collection creation request is sent."""
        # 1. Addressbook with contacts must be synced manually
        data = {"username": self.user.username, "password": "toto"}
        self.client.post(reverse("core:login"), data)
        self.enable_cdav_sync()
        self.addressbook.refresh_from_db()
        self.assertIs(self.addressbook.last_sync, None)

        # 2. Addressbook with no contacts can be considered synced
        user = core_models.User.objects.get(username="user@test2.com")
        abook = user.addressbook_set.first()
        data = {"username": user.username, "password": "toto"}
        self.client.post(reverse("core:login"), data)
        abook.refresh_from_db()
        self.assertIs(abook.last_sync, None)
        # Now enable sync.
        self.enable_cdav_sync()
        abook.refresh_from_db()
        self.assertIsNot(abook.last_sync, None)

    def test_index(self):
        """Test index view."""
        url = reverse("modoboa_contacts:index")
        with self.assertRaises(IOError):
            response = self.client.get(url)
        # self.assertContains(response, '<div id="app">')


class AddressBookViewSetTestCase(TestDataMixin, ModoAPITestCase):
    """Address book ViewSet tests."""

    def test_default(self):
        """Test default endpoint."""
        response = self.client.get(reverse("api:addressbook-default"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["name"], "Contacts")

    def test_sync_to_cdav(self):
        """Test sync to CardDAV endpoint."""
        data = {"username": self.user.username, "password": "toto"}
        response = self.client.post(reverse("core:login"), data)
        self.enable_cdav_sync()
        with httmock.HTTMock(
                mocks.options_mock,
                mocks.mkcol_mock,
                mocks.put_mock,
                mocks.propfind_mock):
            response = self.client.get(reverse("api:addressbook-sync-to-cdav"))
        self.assertEqual(response.status_code, 200)
        abook = self.user.addressbook_set.first()
        self.assertIsNot(abook.sync_token, None)
        contact = abook.contact_set.first()
        self.assertIsNot(contact.etag, None)

    def test_sync_from_cdav(self):
        """Test sync from CardDAV endpoint."""
        data = {"username": self.user.username, "password": "toto"}
        response = self.client.post(reverse("core:login"), data)
        self.enable_cdav_sync()
        self.user.addressbook_set.update(last_sync=timezone.now())
        with httmock.HTTMock(
                mocks.options_mock, mocks.report_mock, mocks.get_mock):
            response = self.client.get(
                reverse("api:addressbook-sync-from-cdav"))
        self.assertEqual(response.status_code, 200)


class CategoryViewSetTestCase(TestDataMixin, ModoAPITestCase):
    """Category ViewSet tests."""

    def test_get_categories(self):
        """Check category list endpoint."""
        url = reverse("api:category-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)

    def test_create_category(self):
        """Create a new category."""
        url = reverse("api:category-list")
        data = {"name": u"Amitiés"}
        response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, 201)

    def test_update_category(self):
        """Update a category."""
        url = reverse("api:category-detail", args=[self.category.pk])
        data = {"name": u"Test"}
        response = self.client.put(url, data, format="json")
        self.assertEqual(response.status_code, 200)
        self.category.refresh_from_db()
        self.assertEqual(self.category.name, "Test")

    def test_delete_category(self):
        """Try to delete a category."""
        url = reverse("api:category-detail", args=[self.category.pk])
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 204)
        with self.assertRaises(models.Category.DoesNotExist):
            self.category.refresh_from_db()


class ContactViewSetTestCase(TestDataMixin, ModoAPITestCase):
    """Contact ViewSet tests."""

    def test_contact_list(self):
        """Check contact list endpoint."""
        url = reverse("api:contact-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 3)

        response = self.client.get("{}?search=homer".format(url))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)

    def test_filter_by_category(self):
        """Check filtering."""
        url = reverse("api:contact-list")
        response = self.client.get(
            "{}?category={}".format(url, self.category.name))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)

    def test_create_contact(self):
        """Create a new contact."""
        data = {"username": self.user.username, "password": "toto"}
        response = self.client.post(reverse("core:login"), data)
        self.enable_cdav_sync()
        self.user.addressbook_set.update(last_sync=timezone.now())
        data = {
            "first_name": "Magie", "last_name": "Simpson",
            "emails": [
                {"address": "magie@simpson.com", "type": "home"}
            ],
            "phone_numbers": [
                {"number": "+33123456789", "type": "home"}
            ]
        }
        url = reverse("api:contact-list")
        with httmock.HTTMock(mocks.options_mock, mocks.put_mock):
            response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, 201)
        contact = models.Contact.objects.get(pk=response.data["pk"])
        self.assertEqual(contact.emails.first().address, "magie@simpson.com")
        self.assertEqual(
            contact.phone_numbers.first().number,
            response.data["phone_numbers"][0]["number"])
        self.assertEqual(contact.display_name, "Magie Simpson")
        self.assertIsNot(contact.etag, None)

    def test_create_contact_quick(self):
        """Create a contact with minimal information."""
        data = {
            "emails": [
                {"address": "magie@simpson.com", "type": "home"}
            ]
        }
        url = reverse("api:contact-list")
        response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data["display_name"][0],
                         "Name or display name required")

        data["display_name"] = "Magie Simpson"
        response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, 201)

    def test_create_contact_with_category(self):
        """Create a new contact with a category."""
        data = {
            "first_name": "Magie", "last_name": "Simpson",
            "emails": [
                {"address": "magie@simpson.com", "type": "home"}
            ],
            "categories": [self.category.pk]
        }
        url = reverse("api:contact-list")
        response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, 201)
        contact = models.Contact.objects.get(pk=response.data["pk"])
        self.assertEqual(contact.categories.first(), self.category)

    def test_update_contact(self):
        """Update existing contact."""
        data = {"username": self.user.username, "password": "toto"}
        self.client.post(reverse("core:login"), data)
        self.enable_cdav_sync()
        self.user.addressbook_set.update(last_sync=timezone.now())
        url = reverse("api:contact-detail", args=[self.contact.pk])
        email_pk = self.contact.emails.first().pk
        data = {
            "first_name": "Homer 1", "last_name": "Simpson",
            "emails": [
                {"address": "duff@simpson.com", "type": "work"},
                {"address": "homer@simpson.com", "type": "other"}
            ],
            "phone_numbers": [
                {"number": "+33123456789", "type": "home"},
                {"number": "01234567889", "type": "work"},
            ],
            "categories": [self.category.pk]
        }
        with httmock.HTTMock(
                mocks.options_mock, mocks.put_mock, mocks.report_mock):
            response = self.client.put(url, data, format="json")
        self.assertEqual(response.status_code, 200)

        self.contact.refresh_from_db()
        self.assertEqual(self.contact.first_name, "Homer 1")
        self.assertEqual(self.contact.emails.count(), 2)
        self.assertEqual(
            models.EmailAddress.objects.get(pk=email_pk).type, "other")
        self.assertEqual(self.contact.phone_numbers.count(), 2)
        self.assertEqual(self.contact.categories.first(), self.category)
        self.assertIsNot(self.contact.etag, None)

        data["emails"].pop(1)
        data["phone_numbers"].pop(1)
        with httmock.HTTMock(
                mocks.options_mock, mocks.put_mock, mocks.report_mock):
            response = self.client.put(url, data, format="json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.contact.emails.count(), 1)
        self.assertEqual(self.contact.phone_numbers.count(), 1)

    def test_delete_contact(self):
        """Try to delete a contact."""
        data = {"username": self.user.username, "password": "toto"}
        response = self.client.post(reverse("core:login"), data)
        self.enable_cdav_sync()
        url = reverse("api:contact-detail", args=[self.contact.pk])
        with httmock.HTTMock(mocks.options_mock, mocks.delete_mock):
            response = self.client.delete(url)
        self.assertEqual(response.status_code, 204)
        with self.assertRaises(models.Contact.DoesNotExist):
            self.contact.refresh_from_db()


class EmailAddressViewSetTestCase(TestDataMixin, ModoAPITestCase):
    """EmailAddressViewSet tests."""

    def test_emails_list(self):
        """Check list endpoint."""
        url = reverse("api:emailaddress-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 3)

        response = self.client.get("{}?search=homer".format(url))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)

        response = self.client.get("{}?search=Marge".format(url))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)

        response = self.client.get("{}?search=Simpson".format(url))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 3)



"""Declare and register the contacts extension."""

from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy

from modoboa.core.extensions import ModoExtension, exts_pool
from modoboa.parameters import tools as param_tools

from . import __version__
from . import forms


class Contacts(ModoExtension):
    """Plugin declaration."""

    name = "modoboa_contacts"
    label = ugettext_lazy("Contacts")
    version = __version__
    description = ugettext_lazy("Address book")
    url = "contacts"
    topredirection_url = reverse_lazy("modoboa_contacts:index")

    def load(self):
        param_tools.registry.add(
            "user", forms.UserSettings, ugettext_lazy("Contacts"))


exts_pool.register_extension(Contacts)



"""Contacts models."""

from __future__ import unicode_literals

import os
import uuid

from dateutil.parser import parse
import vobject

from django.db import models
from django.utils.translation import ugettext as _

from modoboa.lib import exceptions as lib_exceptions
from modoboa.parameters import tools as param_tools

from . import constants
from . import __version__


class AddressBook(models.Model):
    """An address book."""

    name = models.CharField(max_length=50)
    sync_token = models.TextField(blank=True)
    last_sync = models.DateTimeField(null=True)
    user = models.ForeignKey("core.User", on_delete=models.CASCADE)
    _path = models.TextField()

    @property
    def url(self):
        server_location = param_tools.get_global_parameter(
            "server_location", app="modoboa_radicale")
        if not server_location:
            raise lib_exceptions.InternalError(
                _("Server location is not set, please fix it."))
        return os.path.join(server_location, self.user.username, self._path)


class Category(models.Model):
    """A category for contacts."""

    user = models.ForeignKey("core.User", on_delete=models.CASCADE)
    name = models.CharField(max_length=50)


class Contact(models.Model):
    """A contact."""

    addressbook = models.ForeignKey(AddressBook, on_delete=models.CASCADE)
    uid = models.CharField(
        max_length=100, unique=True, null=True, db_index=True)
    # Can't define an index on etag field because of MySQL...
    etag = models.TextField(blank=True)
    first_name = models.CharField(max_length=30, blank=True)
    last_name = models.CharField(max_length=30, blank=True)
    display_name = models.CharField(max_length=60, blank=True)
    birth_date = models.DateField(null=True)

    company = models.CharField(max_length=100, blank=True)
    position = models.CharField(max_length=200, blank=True)

    address = models.CharField(max_length=200, blank=True)
    zipcode = models.CharField(max_length=15, blank=True)
    city = models.CharField(max_length=100, blank=True)
    country = models.CharField(max_length=100, blank=True)
    state = models.CharField(max_length=100, blank=True)

    note = models.TextField(blank=True)

    categories = models.ManyToManyField(Category, blank=True)

    def __init__(self, *args, **kwargs):
        """Set uid for new object."""
        super(Contact, self).__init__(*args, **kwargs)
        if not self.pk:
            self.uid = "{}.vcf".format(uuid.uuid4())

    @property
    def url(self):
        return "{}/{}.vcf".format(self.addressbook.url, self.uid)

    def to_vcard(self):
        """Convert this contact to a vCard."""
        card = vobject.vCard()
        card.add("prodid").value = "-//Modoboa//Contacts plugin {}//EN".format(
            __version__)
        card.add("uid").value = self.uid
        card.add("n").value = vobject.vcard.Name(
            family=self.last_name, given=self.first_name)
        card.add("fn").value = self.display_name
        card.add("org").value = [self.company]
        card.add("title").value = self.position
        card.add("adr").value = vobject.vcard.Address(
            street=self.address, city=self.city, code=self.zipcode,
            country=self.country, region=self.state)
        if self.birth_date:
            card.add("bday").value = self.birth_date.isoformat()
        card.add("note").value = self.note
        for email in EmailAddress.objects.filter(contact=self):
            attr = card.add("email")
            attr.value = email.address
            attr.type_param = email.type
        for phone in PhoneNumber.objects.filter(contact=self):
            attr = card.add("tel")
            attr.value = phone.number
            attr.type_param = phone.type
        return card.serialize()

    def update_from_vcard(self, content):
        """Update this contact according to given vcard."""
        vcard = vobject.readOne(content)
        self.uid = vcard.uid.value
        name = getattr(vcard, "n", None)
        if name:
            self.first_name = name.value.given
            self.last_name = name.value.family
        address = getattr(vcard, "adr", None)
        if address:
            self.address = address.value.street
            self.zipcode = address.value.code
            self.city = address.value.city
            self.state = address.value.region
            self.country = address.value.country
        birth_date = getattr(vcard, "bday", None)
        if birth_date:
            self.birth_date = parse(birth_date.value)
        for cfield, mfield in constants.CDAV_TO_MODEL_FIELDS_MAP.items():
            value = getattr(vcard, cfield, None)
            if value:
                if isinstance(value.value, list):
                    setattr(self, mfield, value.value[0])
                else:
                    setattr(self, mfield, value.value)
        self.save()
        email_list = getattr(vcard, "email_list", [])
        EmailAddress.objects.filter(contact=self).delete()
        to_create = []
        for email in email_list:
            to_create.append(EmailAddress(
                contact=self, address=email.value.lower(),
                type=email.type_param.lower()
            ))
        EmailAddress.objects.bulk_create(to_create)
        PhoneNumber.objects.filter(contact=self).delete()
        to_create = []
        phone_list = getattr(vcard, "tel_list", [])
        for tel in phone_list:
            to_create.append(PhoneNumber(
                contact=self, number=tel.value.lower(),
                type=tel.type_param.lower()
            ))
        PhoneNumber.objects.bulk_create(to_create)


class EmailAddress(models.Model):
    """An email address."""

    contact = models.ForeignKey(Contact, related_name="emails",
                                on_delete=models.CASCADE)
    address = models.EmailField()
    type = models.CharField(
        max_length=20, choices=constants.EMAIL_TYPES)


class PhoneNumber(models.Model):
    """A phone number."""

    contact = models.ForeignKey(Contact, related_name="phone_numbers",
                                on_delete=models.CASCADE)
    number = models.CharField(max_length=40)
    type = models.CharField(
        max_length=20, choices=constants.PHONE_TYPES)

    def __str__(self):
        return "{}: {}".format(self.type, self.number)



"""Contacts API urls."""

from rest_framework import routers

from . import viewsets


router = routers.SimpleRouter()
router.register(
    r"address-books", viewsets.AddressBookViewSet, base_name="addressbook")
router.register(r"categories", viewsets.CategoryViewSet, base_name="category")
router.register(r"contacts", viewsets.ContactViewSet, base_name="contact")
router.register(
    r"emails", viewsets.EmailAddressViewSet, base_name="emailaddress")

urlpatterns = router.urls



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-04-20 11:33
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Contact',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('first_name', models.CharField(blank=True, max_length=30)),
                ('last_name', models.CharField(blank=True, max_length=30)),
                ('display_name', models.CharField(blank=True, max_length=60)),
                ('birth_date', models.DateField(null=True)),
                ('company', models.CharField(blank=True, max_length=100)),
                ('position', models.CharField(blank=True, max_length=200)),
                ('address', models.CharField(blank=True, max_length=200)),
                ('zipcode', models.CharField(blank=True, max_length=15)),
                ('city', models.CharField(blank=True, max_length=100)),
                ('country', models.CharField(blank=True, max_length=100)),
                ('state', models.CharField(blank=True, max_length=100)),
                ('note', models.TextField(blank=True)),
                ('categories', models.ManyToManyField(blank=True, to='modoboa_contacts.Category')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='EmailAddress',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('address', models.EmailField(max_length=254)),
                ('type', models.CharField(choices=[(b'home', 'Home'), (b'work', 'Work'), (b'other', 'Other')], max_length=20)),
                ('contact', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='emails', to='modoboa_contacts.Contact')),
            ],
        ),
        migrations.CreateModel(
            name='PhoneNumber',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('number', models.CharField(max_length=40)),
                ('type', models.CharField(choices=[(b'home', 'Home'), (b'work', 'Work'), (b'other', 'Other'), (b'main', 'Main'), (b'cellular', 'Cellular'), (b'fax', 'Fax'), (b'pager', 'Pager')], max_length=20)),
                ('contact', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='phone_numbers', to='modoboa_contacts.Contact')),
            ],
        ),
    ]



"""Contacts forms."""

from django import forms
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from modoboa.lib import cryptutils
from modoboa.lib import form_utils
from modoboa.lib import signals as lib_signals
from modoboa.parameters import forms as param_forms

from . import tasks


class UserSettings(param_forms.UserParametersForm):
    """User settings."""

    app = "modoboa_contacts"

    sep1 = form_utils.SeparatorField(label=_("Synchronization"))

    enable_carddav_sync = form_utils.YesNoField(
        initial=False,
        label=_("Synchonize address book using CardDAV"),
        help_text=_(
            "Choose to synchronize or not your address book using CardDAV. "
            "You will be able to access your contacts from the outside."
        )
    )

    sync_frequency = forms.IntegerField(
        initial=300,
        label=_("Synchronization frequency"),
        help_text=_(
            "Interval in seconds between 2 synchronization requests"
        )
    )

    visibility_rules = {
        "sync_frequency": "enable_carddav_sync=True"
    }

    def clean_sync_frequency(self):
        """Make sure frequency is a positive integer."""
        if self.cleaned_data["sync_frequency"] < 60:
            raise forms.ValidationError(
                _("Minimum allowed value is 60s")
            )
        return self.cleaned_data["sync_frequency"]

    def save(self, *args, **kwargs):
        """Create remote cal if necessary."""
        super(UserSettings, self).save(*args, **kwargs)
        if not self.cleaned_data["enable_carddav_sync"]:
            return
        abook = self.user.addressbook_set.first()
        if abook.last_sync:
            return
        request = lib_signals.get_request()
        tasks.create_cdav_addressbook(
            abook, cryptutils.decrypt(request.session["password"]))
        if not abook.contact_set.exists():
            abook.last_sync = timezone.now()
            abook.save(update_fields=["last_sync"])



"""Async tasks."""

from django.utils import timezone

from modoboa.lib import cryptutils

from .lib import carddav
from . import models


def get_cdav_client(request, addressbook, write_support=False):
    """Instantiate a new CardDAV client."""
    return carddav.PyCardDAV(
        addressbook.url, user=request.user.username,
        passwd=cryptutils.decrypt(request.session["password"]),
        write_support=write_support
    )


def create_cdav_addressbook(addressbook, password):
    """Create CardDAV address book."""
    clt = carddav.PyCardDAV(
        addressbook.url, user=addressbook.user.username,
        passwd=password,
        write_support=True
    )
    clt.create_abook()


def push_addressbook_to_carddav(request, addressbook):
    """Push every addressbook item to carddav collection.

    Use only once.
    """
    clt = get_cdav_client(request, addressbook, True)
    for contact in addressbook.contact_set.all():
        href, etag = clt.upload_new_card(contact.uid, contact.to_vcard())
        contact.etag = etag
        contact.save(update_fields=["etag"])
    addressbook.last_sync = timezone.now()
    addressbook.sync_token = clt.get_sync_token()
    addressbook.save(update_fields=["last_sync", "sync_token"])


def sync_addressbook_from_cdav(request, addressbook):
    """Fetch changes from CardDAV server."""
    clt = get_cdav_client(request, addressbook)
    changes = clt.sync_vcards(addressbook.sync_token)
    if not len(changes["cards"]):
        return
    for card in changes["cards"]:
        # UID sometimes embded .vcf extension, sometimes not...
        long_uid = card["href"].split("/")[-1]
        short_uid = long_uid.split(".")[0]
        if "200" in card["status"]:
            content = clt.get_vcard(card["href"]).decode()
            contact = models.Contact.objects.filter(
                uid__in=[long_uid, short_uid]).first()
            if not contact:
                contact = models.Contact(addressbook=addressbook)
            if contact.etag != card["etag"]:
                contact.etag = card["etag"]
                contact.update_from_vcard(content)
        elif "404" in card["status"]:
            models.Contact.objects.filter(
                uid__in=[long_uid, short_uid]).delete()
    addressbook.last_sync = timezone.now()
    addressbook.sync_token = changes["token"]
    addressbook.save(update_fields=["last_sync", "sync_token"])


def push_contact_to_cdav(request, contact):
    """Upload new contact to cdav collection."""
    clt = get_cdav_client(request, contact.addressbook, True)
    path, etag = clt.upload_new_card(contact.uid, contact.to_vcard())
    contact.etag = etag
    contact.save(update_fields=["etag"])


def update_contact_cdav(request, contact):
    """Update existing contact."""
    clt = get_cdav_client(request, contact.addressbook, True)
    uid = contact.uid
    if not uid.endswith(".vcf"):
        uid += ".vcf"
    result = clt.update_vcard(contact.to_vcard(), uid, contact.etag)
    contact.etag = result["cards"][0]["etag"]
    contact.save(update_fields=["etag"])


def delete_contact_cdav(request, contact):
    """Delete a contact."""
    clt = get_cdav_client(request, contact.addressbook, True)
    uid = contact.uid
    if not uid.endswith(".vcf"):
        uid += ".vcf"
    clt.delete_vcard(uid, contact.etag)



"""Contacts constants."""

from django.utils.translation import ugettext_lazy

EMAIL_TYPES = (
    ("home", ugettext_lazy("Home")),
    ("work", ugettext_lazy("Work")),
    ("other", ugettext_lazy("Other"))
)

PHONE_TYPES = (
    ("home", ugettext_lazy("Home")),
    ("work", ugettext_lazy("Work")),
    ("other", ugettext_lazy("Other")),
    ("main", ugettext_lazy("Main")),
    ("cellular", ugettext_lazy("Cellular")),
    ("fax", ugettext_lazy("Fax")),
    ("pager", ugettext_lazy("Pager"))
)

CDAV_TO_MODEL_FIELDS_MAP = {
    "org": "company",
    "title": "position",
    "note": "note",
}



"""Webmail handlers."""

from django.db.models import signals
from django.urls import reverse
from django.dispatch import receiver
from django.utils.translation import ugettext as _

from modoboa.admin import models as admin_models
from modoboa.core import signals as core_signals
from modoboa.lib import signals as lib_signals

from . import models


@receiver(core_signals.extra_user_menu_entries)
def menu(sender, location, user, **kwargs):
    """Return extra menu entry."""
    if location != "top_menu" or not hasattr(user, "mailbox"):
        return []
    return [
        {"name": "contacts",
         "label": _("Contacts"),
         "url": reverse("modoboa_contacts:index")},
    ]


@receiver(core_signals.get_top_notifications)
def check_addressbook_first_sync(sender, include_all, **kwargs):
    """Check if address book first sync has been made."""
    request = lib_signals.get_request()
    qset = (
        request.user.addressbook_set.filter(last_sync__isnull=True)
    )
    condition = (
        not request.user.parameters.get_value("enable_carddav_sync") or
        not qset.exists()
    )
    if condition:
        return []
    return [{
        "id": "abook_sync_required",
        "url": reverse("modoboa_contacts:index"),
        "text": _("Your address book must be synced"),
        "level": "warning"
    }]


@receiver(signals.post_save, sender=admin_models.Mailbox)
def create_addressbook(sender, instance, created, **kwargs):
    """Create default address book for new mailbox."""
    if not created:
        return
    models.AddressBook.objects.create(
        user=instance.user, name="Contacts", _path="contacts")


@receiver(core_signals.extra_static_content)
def inject_sync_poller(sender, caller, st_type, user, **kwargs):
    """Inject javascript code."""
    condition = (
        caller != "top" or
        st_type != "js" or
        not hasattr(user, "mailbox") or
        not user.parameters.get_value("enable_carddav_sync")
    )
    if condition:
        return ""
    return """<script>
$(document).ready(function () {
    new Poller('%s', {
        interval: %d * 1000
    });
});
</script>
""" % (reverse("api:addressbook-sync-from-cdav"),
       user.parameters.get_value("sync_frequency"))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-10-05 12:15
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('modoboa_contacts', '0002_auto_20180124_2311'),
    ]

    operations = [
        migrations.CreateModel(
            name='AddressBook',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('sync_token', models.TextField(blank=True)),
                ('last_sync', models.DateTimeField(null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('_path', models.TextField()),
            ],
        ),
        migrations.AddField(
            model_name='contact',
            name='uid',
            field=models.CharField(max_length=100, null=True, unique=True, db_index=True),
        ),
        migrations.AddField(
            model_name='contact',
            name='etag',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='contact',
            name='addressbook',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='modoboa_contacts.AddressBook'),
        ),
    ]



"""Contacts views."""

from django.views import generic

from django.contrib.auth import mixins as auth_mixins


class IndexView(auth_mixins.LoginRequiredMixin, generic.TemplateView):
    """Simple view to display index page."""

    template_name = "modoboa_contacts/index.html"

    def get_context_data(self, **kwargs):
        """Set menu selection."""
        context = super(IndexView, self).get_context_data(**kwargs)
        context.update({
            "selection": "contacts",
            "sync_enabled": self.request.user.parameters.get_value(
                "enable_carddav_sync"),
            "abook_synced": self.request.user.addressbook_set.filter(
                last_sync__isnull=False).exists()
        })
        return context



"""Contacts serializers."""

from django.utils.translation import ugettext as _

from rest_framework import serializers

from . import models
from . import tasks


class AddressBookSerializer(serializers.ModelSerializer):
    """Address book serializer."""

    class Meta:
        model = models.AddressBook
        fields = ("pk", "name", "url")


class EmailAddressSerializer(serializers.ModelSerializer):
    """Email address serializer."""

    class Meta:
        model = models.EmailAddress
        fields = ("pk", "address", "type")


class EmailAddressWithNameSerializer(serializers.ModelSerializer):
    """Email address + contact name serializer."""

    display_name = serializers.SerializerMethodField()

    class Meta:
        model = models.EmailAddress
        fields = ("pk", "address", "type", "display_name")

    def get_display_name(self, obj):
        """Return display name."""
        if obj.contact.display_name:
            return obj.contact.display_name
        return u"{} {}".format(obj.contact.first_name, obj.contact.last_name)


class PhoneNumberSerializer(serializers.ModelSerializer):
    """Phone number serializer."""

    class Meta:
        model = models.PhoneNumber
        fields = ("pk", "number", "type")


class CategorySerializer(serializers.ModelSerializer):
    """Serializer for Category."""

    class Meta:
        model = models.Category
        fields = ("pk", "name")

    def create(self, validated_data):
        """Use current user."""
        user = self.context["request"].user
        return models.Category.objects.create(user=user, **validated_data)


class ContactSerializer(serializers.ModelSerializer):
    """Contact serializer."""

    emails = EmailAddressSerializer(many=True)
    phone_numbers = PhoneNumberSerializer(many=True, required=False)

    class Meta:
        model = models.Contact
        fields = (
            "pk", "first_name", "last_name", "categories", "emails",
            "phone_numbers", "company", "position",
            "address", "zipcode", "city", "country", "state",
            "note", "birth_date", "display_name"
        )

    def validate(self, data):
        """Make sure display name or first/last names are set."""
        condition = (
            not data.get("first_name") and
            not data.get("last_name") and
            not data.get("display_name")
        )
        if condition:
            msg = _("Name or display name required")
            raise serializers.ValidationError({
                "first_name": msg,
                "last_name": msg,
                "display_name": msg
            })
        if not data.get("display_name"):
            data["display_name"] = data.get("first_name", "")
            if data["display_name"]:
                data["display_name"] += " "
            data["display_name"] += data.get("last_name", "")
        return data

    def create(self, validated_data):
        """Use current user."""
        request = self.context["request"]
        addressbook = request.user.addressbook_set.first()
        categories = validated_data.pop("categories", [])
        emails = validated_data.pop("emails")
        phone_numbers = validated_data.pop("phone_numbers", [])
        contact = models.Contact.objects.create(
            addressbook=addressbook, **validated_data)
        to_create = []
        for email in emails:
            to_create.append(models.EmailAddress(contact=contact, **email))
        models.EmailAddress.objects.bulk_create(to_create)
        to_create = []
        for phone_number in phone_numbers:
            to_create.append(
                models.PhoneNumber(contact=contact, **phone_number))
        if to_create:
            models.PhoneNumber.objects.bulk_create(to_create)
        if categories:
            for category in categories:
                contact.categories.add(category)
        condition = (
            addressbook.last_sync and
            addressbook.user.parameters.get_value("enable_carddav_sync")
        )
        if condition:
            tasks.push_contact_to_cdav(request, contact)
        return contact

    def update_emails(self, instance, emails):
        """Update instance emails."""
        local_addresses = []
        local_objects = []
        for email in instance.emails.all():
            local_addresses.append(email.address)
            local_objects.append(email)
        to_create = []
        for email in emails:
            if email["address"] not in local_addresses:
                to_create.append(
                    models.EmailAddress(contact=instance, **email))
                continue
            index = local_addresses.index(email["address"])
            local_email = local_objects[index]
            condition = (
                local_email.type != email["type"] or
                local_email.address != email["address"])
            if condition:
                local_email.type = email["type"]
                local_email.address = email["address"]
                local_email.save()
            local_addresses.pop(index)
            local_objects.pop(index)
        models.EmailAddress.objects.filter(
            pk__in=[email.pk for email in local_objects]).delete()
        models.EmailAddress.objects.bulk_create(to_create)

    def update_phone_numbers(self, instance, phone_numbers):
        """Update instance phone numbers."""
        local_phones = []
        local_objects = []
        for phone in instance.phone_numbers.all():
            local_phones.append(phone.number)
            local_objects.append(phone)
        to_create = []
        for phone in phone_numbers:
            if phone["number"] not in local_phones:
                to_create.append(
                    models.PhoneNumber(contact=instance, **phone))
                continue
            index = local_phones.index(phone["number"])
            local_phone = local_objects[index]
            condition = (
                local_phone.type != phone["type"] or
                local_phone.number != phone["number"])
            if condition:
                local_phone.type = phone["type"]
                local_phone.number = phone["number"]
                local_phone.save()
            local_phones.pop(index)
            local_objects.pop(index)
        instance.phone_numbers.filter(
            pk__in=[phone.pk for phone in local_objects]).delete()
        models.PhoneNumber.objects.bulk_create(to_create)

    def update(self, instance, validated_data):
        """Update contact."""
        emails = validated_data.pop("emails")
        phone_numbers = validated_data.pop("phone_numbers", [])
        for key, value in validated_data.items():
            setattr(instance, key, value)
        instance.save()

        self.update_emails(instance, emails)
        self.update_phone_numbers(instance, phone_numbers)

        condition = (
            instance.addressbook.last_sync and
            instance.addressbook.user.parameters.get_value(
                "enable_carddav_sync")
        )
        if condition:
            tasks.update_contact_cdav(self.context["request"], instance)

        return instance



"""Mocks to test the contacts plugin."""

import httmock


@httmock.urlmatch(method="OPTIONS")
def options_mock(url, request):
    """Simulate options request."""
    return {"status_code": 200, "headers": {"DAV": "addressbook"}}


@httmock.urlmatch(method="MKCOL")
def mkcol_mock(url, request):
    """Simulate collection creation."""
    return {"status_code": 201}


@httmock.urlmatch(method="DELETE")
def delete_mock(url, request):
    """Simulate a DELETE request."""
    return {"status_code": 204}


@httmock.urlmatch(method="PUT")
def put_mock(url, request):
    """Simulate a PUT request."""
    return {"status_code": 200, "headers": {"etag": '"12345"'}}


@httmock.urlmatch(method="PROPFIND")
def propfind_mock(url, request):
    """Simulate a PROPFIND request."""
    content = b"""
<d:multistatus xmlns:d="DAV:" xmlns:cs="http://calendarserver.org/ns/">
    <d:response>
        <d:href>/radicale/user@test.com/contacts/</d:href>
        <d:propstat>
            <d:prop>
                <d:sync-token>http://modoboa.org/ns/sync-token/3145</d:sync-token>
            </d:prop>
            <d:status>HTTP/1.1 200 OK</d:status>
        </d:propstat>
    </d:response>
</d:multistatus>
"""
    return {"status_code": 200, "content": content}


@httmock.urlmatch(method="REPORT")
def report_mock(url, request):
    """Simulate a REPORT request."""
    if url.path.endswith(".vcf"):
        content = b"""
<d:multistatus xmlns:d="DAV:">
    <d:response>
        <d:href>/radicale/user@test.com/contacts/newcard.vcf</d:href>
        <d:propstat>
            <d:prop>
                <d:getetag>"33441-34321"</d:getetag>
            </d:prop>
            <d:status>HTTP/1.1 200 OK</d:status>
        </d:propstat>
    </d:response>
 </d:multistatus>
"""
    else:
        content = b"""
<d:multistatus xmlns:d="DAV:">
    <d:response>
        <d:href>/radicale/user@test.com/contacts/newcard.vcf</d:href>
        <d:propstat>
            <d:prop>
                <d:getetag>"33441-34321"</d:getetag>
            </d:prop>
            <d:status>HTTP/1.1 200 OK</d:status>
        </d:propstat>
    </d:response>
    <d:response>
        <d:href>/radicale/user@test.com/contacts/updatedcard.vcf</d:href>
        <d:propstat>
            <d:prop>
                <d:getetag>"33541-34696"</d:getetag>
            </d:prop>
            <d:status>HTTP/1.1 200 OK</d:status>
        </d:propstat>
    </d:response>
    <d:response>
        <d:href>/radicale/user@test.com/contacts/deletedcard.vcf</d:href>
        <d:status>HTTP/1.1 404 Not Found</d:status>
    </d:response>
    <d:sync-token>http://modoboa.org/ns/sync/5001</d:sync-token>
 </d:multistatus>
"""
    return {"status_code": 200, "content": content}


@httmock.urlmatch(method="GET")
def get_mock(url, request):
    """Simulate a GET request."""
    uid = url.path.split("/")[-1]
    content = """
BEGIN:VCARD
VERSION:3.0
UID:{}
N:Gump;Forrest
FN:Forrest Gump
ORG:Bubba Gump Shrimp Co.
TITLE:Shrimp Man
TEL;TYPE=WORK;VOICE:(111) 555-1212
TEL;TYPE=HOME;VOICE:(404) 555-1212
ADR;TYPE=HOME:;;42 Plantation St.;Baytown;LA;30314;United States of America
EMAIL;TYPE=PREF,INTERNET:forrestgump@example.com
END:VCARD
""".format(uid)
    return {"status_code": 200, "content": content}



# -*- coding: utf-8 -*-
# Generated by Django 1.11.8 on 2018-01-24 23:11
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('modoboa_contacts', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='emailaddress',
            name='type',
            field=models.CharField(choices=[('home', 'Home'), ('work', 'Work'), ('other', 'Other')], max_length=20),
        ),
        migrations.AlterField(
            model_name='phonenumber',
            name='type',
            field=models.CharField(choices=[('home', 'Home'), ('work', 'Work'), ('other', 'Other'), ('main', 'Main'), ('cellular', 'Cellular'), ('fax', 'Fax'), ('pager', 'Pager')], max_length=20),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-10-05 12:15
from __future__ import unicode_literals

import uuid

from django.db import migrations


def create_address_books(apps, schema_editor):
    """Create address books for every account."""
    User = apps.get_model("core", "User")
    AddressBook = apps.get_model("modoboa_contacts", "AddressBook")
    for user in User.objects.filter(mailbox__isnull=False):
        abook = AddressBook.objects.create(
            name="Contacts", user=user, _path="contacts")
        for contact in user.contact_set.all():
            contact.addressbook = abook
            contact.uid = "{}.vcf".format(uuid.uuid4())
            contact.save()


def backward(apps, schema_editor):
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('modoboa_contacts', '0003_auto_20181005_1415'),
    ]

    operations = [
        migrations.RunPython(create_address_books, backward)
    ]
